const St = imports.gi.St;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const Lang = imports.lang;
const Shell = imports.gi.Shell;
const Mainloop = imports.mainloop;
const AppDisplay = imports.ui.appDisplay;
const Clutter = imports.gi.Clutter;

const Extension = imports.misc.extensionUtils.getCurrentExtension();
const Windows = Extension.imports.windows;

const Favorite = Extension.imports.domains.favorite;
const Task = Extension.imports.domains.task;
const Preview = Extension.imports.domains.preview;

const AppFavorites = imports.ui.appFavorites;

const THUMBNAIL_SIZE = 300;

function getThumbnail(window, size) {
	let thumbnail = null;
	
	let mutterWindow = window.get_compositor_private();
	
	if (mutterWindow)
	{
		let windowTexture = mutterWindow.get_texture();
		let [width, height] = windowTexture.get_size();
		let scale = Math.min(1.0, size / width, size / height);
		
		thumbnail = new Clutter.Clone ({ source: windowTexture, reactive: true, width: width * scale, height: height * scale });
	}
	
	return thumbnail;
}

function onClickTaskButton(button, pspec, window)
{
	//numButton is clicked button (left,middle,right)
	let numButton = pspec.get_button();
	
	if(numButton==1) {
		this.tasks.forEach(
			function(task)
			{
				if (task.getWindow() == window)
				{
					if (! task.getWindow().has_focus())
					{
						task.getWindow().activate(global.get_current_time());
						task.getButton().add_style_pseudo_class('active-task');
					}
					else
					{
						task.getWindow().minimize(global.get_current_time());
						task.getButton().remove_style_pseudo_class('active-task');
					}
				}
				else
					task.getButton().remove_style_pseudo_class('active-task');
			}
		);
	}
}

const Taskbar = new Lang.Class({
	Name: 'Taskbar',

	Extends: PanelMenu.Button,

	taskbarBox: null,
	tasks: null,
	favorites: null,
	previews: null,
	windows: null,
	timer: null,

	_init: function() {
		this.tasks = new Array();
		this.previews = new Array();
		this.favorites = new Array();
		
		this.taskbarBox = new St.BoxLayout();
		this.addFavorites();
		this.windows = new Windows.Windows(this, this.onWindowsListChange, this.onWindowChange);
	},

	get: function() {
		return this.taskbarBox;
	},

	addFavorites: function() {
		let launchers = global.settings.get_strv(AppFavorites.getAppFavorites().FAVORITE_APPS_KEY);
		
		for ( let i=0; i<launchers.length; ++i ) {
			let app = Shell.AppSystem.get_default().lookup_app(launchers[i]);
			
			if ( app == null ) {
			    continue;
			}
			
			this.addFavorite(app,i);
		}
	},

	addFavorite: function(app,i) {
		let icon = new AppDisplay.AppIcon(app, { setSizeManually: true, showLabel: false });
		icon.icon.setIconSize(22);

		let button = new St.Button({ style_class: "tkb-task-button" });
		button.set_child(icon.actor);
															
		let signal = button.connect("button-press-event", Lang.bind(this,function() {
			app.open_new_window(-1);
		}));
		
		this.taskbarBox.insert_child_at_index(button,i+1);
		
	    this.favorites.push(new Favorite.Favorite(app,button,i,signal));
	},

	refreshFavorites: function(action, appName) {
		for(let i=0;i<this.favorites.length;i++){
			let favorite = this.favorites[i];
			
			if(favorite.getAppName() == appName) {
				if(action == 'add') {
					favorite.getButton().disconnect(favorite.getSignal());
					favorite.getButton().destroy();
				}
				else if(action == 'remove'){
					this.addFavorite(favorite.getApp(),favorite.getPosition());
					this.favorites.splice(i,1);
					break;
				}
			}
		}
	},

	onWindowsListChange: function(windowsList, type, window) {
		switch(type) {
			case 0: //init or workspace change
				this.refreshTasks(windowsList);
				break;
				
			case 1: //add window
				this.addTask(window);
				break;
				
			case 2: //remove window
				this.removeTask(window);
				break;
		}
	},

	onWindowChange: function(window, type) {
		if (type == 0) //Focus
		{
			this.tasks.forEach(
				function(task)
				{
					if(task.getButton()!=null)
					{
						if (task.getWindow() == window)
							task.getButton().add_style_pseudo_class('active-task');
						else
							task.getButton().remove_style_pseudo_class('active-task');
					}
				}
			);
		}
	},

	refreshTasks: function(windowsList) {
		this.clearTasks();

		for (let i = 0; i < windowsList.length; i++) {
			this.addTask(windowsList[i]);
		}
	},

	clearTasks: function() {
		for (let i = this.tasks.length - 1; i>=0; i--)
		{
			let task = tasks[i];
			task.getSignals().forEach(
				function(signal)
				{
					task.getButton().disconnect(signal);
				}
			);
		
			task.getButton().destroy();
			tasks.splice(i, 1);
		}
	},

	getTaskIndex: function(appName) {
		for(let i=0;i<this.tasks.length;i++){
			let task = this.tasks[i];
			if(task.getAppName() == appName)
				return i;
		}
		
		return null;
	},

	getPositionFromFavorite: function(appName) {
		this.favorites.forEach(
			function(favorite) {
				if(favorite.getAppName() == appName)
					return favorite.getPosition();
			}
		);
		
		return null;
	},

	addTask: function(window) {
		let app = Shell.WindowTracker.get_default().get_window_app(window);
		
		let idx = this.getTaskIndex(app.get_name());
		
		if(idx==null) {
			let icon = new AppDisplay.AppIcon(app, { setSizeManually: true, showLabel: false });
			icon.icon.setIconSize(22);
		
			let button = new St.Button({ style_class: "tkb-task-button" });
			button.set_child(icon.actor);
			
			let signals = [
				button.connect("button-press-event", Lang.bind(this, onClickTaskButton, window)),
				button.connect("enter-event", Lang.bind(this, this.showPreviews, window, this.previews)),
				button.connect("leave-event", Lang.bind(this, this.hidePreviews, this.previews))
			];
			
			if (window.has_focus())
				button.add_style_pseudo_class('active-task');
			
			let position = this.getPositionFromFavorite(app.get_name());
			
			if(position!=null)
				this.taskbarBox.insert_child_at_index(button,position+1);
			else
				this.taskbarBox.add_actor(button);
			
			this.refreshFavorites('add',app.get_name());
			
			this.tasks.push(new Task.Task(window, button, signals, app.get_name()));
		
		} else {
			let task = this.tasks[idx];
			this.tasks.push(new Task.Task(window, task.getButton(), task.getSignals(), task.getAppName()));
		}
	},

	countTaskWithName: function(appName) {
		let count = 0;
		tasks.forEach(
			function(task) {
				if(task.getAppName() == appName)
					count++;
			}
		);
		
		return count;
	},

	removeTask: function(window)
	{
		let index = findTaskIndex(window);
		
		let appName = tasks[index].getAppName();
		
		let count = countTaskWithName(appName);
		
		if (index != null && count == 1)
		{
			let task = this.tasks[index];
			
			task.getSignals().forEach(
				function(signal)
				{
					task.getButton().disconnect(signal);
				}
			);
			
			refreshFavorites('remove',appName);
			
			task.getButton().destroy();
			destroyPreviews(this.previews);
			tasks.splice(index, 1);
			
			return true;
		}
		
		return false;
	},

	findTaskIndex: function(window) {
		for (let indexTask in this.tasks)
		{
			let task = this.tasks[indexTask];
			
			if (task.getWindow() == window)
				return indexTask;
		}
		
		return null;
	},

	destroyPreviews: function(previews) {
		for(let i=0;i<previews.length;i++) {
			let preview = previews[i];
			preview.getPreview().disconnect(preview.getSignal());
			preview.getPreview().destroy();
		}
		
		previews = new Array();
		this.previews = new Array();
	},

	showPreviews: function(button, pspec, window, previews)
	{
		//Remove current preview if necessary
		this.destroyPreviews(previews);
		
		let app = Shell.WindowTracker.get_default().get_window_app(window);
		
		let i = 0;
		this.tasks.forEach(
			function(task) {
				if(app!=null && task.getAppName() == app.get_name()) 
				{
					let thumbnail = getThumbnail(window, THUMBNAIL_SIZE);
					
					if(thumbnail!=null) {
						let preview = new St.Button({ style_class: "tkb-preview"});
						let signal = preview.connect("button-press-event", Lang.bind(this, onClickTaskButton, task.getWindow()));
						
						preview.add_actor(thumbnail);
									     
						Main.layoutManager.addChrome(preview);
						
						let [left, top] = button.get_transformed_position();
						
						let [width, height] = thumbnail.get_size();
						
						preview.set_position(left + (width * i) - 10, 22);
						
						previews.push(new Preview.Preview(preview, signal, width, height));
						
						i++;
					}
				}
			}
		);
	},

	hidePreviews: function(button,pspec,previews)
	{
		//Destroy Preview if displaying and mouse not over it
		if (previews.length>0)
		{
			let [x, y, mods] = global.get_pointer();
			
			let [left, top] = previews[0].getPreview().get_position();
			
			let width = this.getPreviewsWidth(previews);
			
			if((x<left || x>left+width || y<top || y>top+previews[0].getHeight()) && y > 24) 
			{
				this.destroyPreviews(previews);
			}
			
			Mainloop.timeout_add(50, Lang.bind(this,this.hidePreviews,button,pspec,previews));
		}
	},

	getPreviewsWidth: function(previews) {
		let width = 0;
		previews.forEach(
			function(preview) {
				width += preview.getWidth();
			}	
		);
		return width;
	},

	stop: function() {
		//remove any possible existing previews
		destroyPreviews(this.previews);
		previews = null;
		
		//remove favorites and disconnect
		favorites.forEach(
			function(favorite) {
				if(favorite!=null){
					favorite.getButton().disconnect(favorite.getSignal());
					favorite.getButton().destroy();
				}
			}
		);
		favorites = null;
		
		//clear taskslist
		clearTasks();
		tasks = null;
		
		//destroy windows
		this.windows.destruct();
		this.windows = null;
		
		//finally remove the taskbar
	    	Main.panel._leftBox.remove_child(this.taskbarBox);
	    	this.taskbarBox = null;
	}
});

let taskbar;

function init() {

}

function enable() {
	taskbar = new Taskbar();
	Main.panel._leftBox.insert_child_at_index(taskbar.get(), 1);	
}

function disable() {
	taskbar.stop();
	taskbar.destroy();
}
